//
// effectSsse3.c
//
// (c)Copyright Spacesoft corp., 2007 All rights reserved.
//                               Hiro KITAYAMA
//
#include <windows.h>
#include <math.h>


#define     SSSE3

//---------------------------------------------------------------------------
//
// GtFNg,SIMD,
//
//      0  -1   0   /
//  (  -1   0   1  /  4  )  + 128
//      0   1   0 /
//
void effectSsse3( LPBYTE pAligned[], LONG width, LONG height, int alpha )
{
    LPBYTE  lpSrc = pAligned[1] ;
    LPBYTE  lpDst = pAligned[0] ;
    int remainY;
    short mulMask0[] ={  0,0, -1,0,  0,0,  0,0};
    short mulMask1[] ={ -1,0,  0,0,  1,0,  0,0};
    short mulMask2[] ={  0,0,  1,0,  0,0,  0,0};
#ifdef  SSSE3
    char  shuffMask[] ={ 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
#endif

    float addConst = 128.0f;

    _asm
    {
        mov         eax, [width]        ; eax = width
        shl         eax, 2              ; eax = width*4, = scan length
        mov         esi, [lpSrc]        ; src 1st line
        lea         ebx, [esi+eax]      ; src 2nd line
        lea         edx, [ebx+eax]      ; src 3rd line

        mov         edi, [lpDst]        ; edi = dst
        lea         edi, [edi+eax+4]    ; edi = dst

        movdqu      xmm7, [mulMask0]    ; xmm7 = mask
        movdqu      xmm6, [mulMask1]    ; xmm6 = mask
        movdqu      xmm5, [mulMask2]    ; xmm5 = mask
        movss       xmm4, [addConst]    ; xmm4 = constant
#ifdef  SSSE3
        movdqu      xmm3, [shuffMask]    ; xmm3 = mask
#endif

        xor         eax, eax            ; index

        mov         ecx, [height]       ; ax = height
        dec         ecx                 ; ax = height - 1
        dec         ecx                 ; ax = height - 2
    loopY:
        mov         [remainY], ecx


        mov         ecx, [width]        ; ecx = width
        dec         ecx                 ; ecx = width - 1
        dec         ecx                 ; ecx = width - 2

    loopX:
        movdqu      xmm0, [esi+eax]     ; 1st line 00M3M3M3 00M2M2M2 00M1M1M1 00M0M0M0
        movdqu      xmm1, [ebx+eax]     ; 2nd line 00M3M3M3 00M2M2M2 00M1M1M1 00M0M0M0
        movdqu      xmm2, [edx+eax]     ; 3rd line 00M3M3M3 00M2M2M2 00M1M1M1 00M0M0M0

        psrld       xmm0, 16+2          ; 1st line 000000M3 000000M2 000000M1 000000M0 / 4
        psrld       xmm1, 16+2          ; 2nd line 000000M3 000000M2 000000M1 000000M0 / 4
        psrld       xmm2, 16+2          ; 3rd line 000000M3 000000M2 000000M1 000000M0 / 4


        pmullw      xmm0, xmm7          ; 1st line 0000 00M3  0000 00M2  0000 00M1  0000 00M0
        pmullw      xmm1, xmm6          ; 2nd line 0000 00M3  0000 00M2  0000 00M1  0000 00M0
        pmullw      xmm2, xmm5          ; 3rd line 0000 00M3  0000 00M2  0000 00M1  0000 00M0

        paddsw      xmm0, xmm1          ; xmm0 =   1st line + 2nd line
        paddsw      xmm0, xmm2          ; xmm0 =   1st line + 2nd line + 3rd line

        cvtdq2ps    xmm0, xmm0          ; (float)
        haddps      xmm0, xmm0          ; xmm0 =   0   M2   M2   M1+M0(float)

        haddps      xmm0, xmm0          ; xmm0 =   0   M2   M2   M2+M1+M0(float)

        addss       xmm0, xmm4



        cvtps2dq    xmm1, xmm0          ; xmm1 =   -------- -------- -------- 0000XXXX(DWORD)


        packuswb    xmm1, xmm1          ; saturate(0-255)

#ifdef  SSSE3
        pshufb      xmm1, xmm3          ; 0-78- 24փRs[
#else
        punpcklbw   xmm1, xmm1          ; 0-78- 24փRs[
                                        ; xmm1 =   -------- -------- -------- 0000XXXX(DWORD)
        movdqa      xmm3, xmm1          ; xmm3 =   -------- -------- -------- 0000XXXX(DWORD)
        pslldq      xmm3, 1             ; xmm3 =   -------- -------- -------- 00XXXX00(DWORD)
        por         xmm1, xmm3          ; xmm3 =   -------- -------- -------- 00XXXXXX(DWORD)
#endif


        movd        [edi+eax], xmm1         ; write result

        lea         eax, [eax+4]
        dec         ecx
        jnz         loopX


        lea         eax, [eax+8]

        mov         ecx, [remainY]
        dec         ecx
        jnz         loopY
    }
}
