//
// effect.c
//
// (c)Copyright Spacesoft corp., 2007 All rights reserved.
//                               Hiro KITAYAMA
//
#include <windows.h>

//---------------------------------------------------------------------------
//
// 32 bit RGB to 32 bit monochrome bmp
//
void toMono32Bmp( LPBYTE pInSrc, LPBYTE pInDst, LONG width, LONG height )
{
    int loopCounter = (height*width)/4;
    float monoConst[]={0.114478f,0.586611f,0.298912f,0.000000f};

    _asm 
    {
        xor         ebx, ebx
        mov         esi, pInSrc         // esi = src
        mov         edi, pInDst         // edi = dst
        mov         ecx, loopCounter    // loop counter


        pxor        xmm0,xmm0           // fixed zero
        movdqu      xmm7, monoConst


                                        // xmm0 = zero(integer)
                                        // xmm2 = source
                                        // xmm5 = destination
                                        // xmm7 = mono constant
    loopLbl:
        movdqa      xmm1, [esi+ebx]     // src
        movdqa      xmm2, xmm1          // save src



                                        // p3, p2=================================
        pshufd      xmm1, xmm2, 0eh     // p3, p2ʂփRs[C----1110b
                                        //
                                        // xmm2:   p3   p2   p1  p0
                                        // xmm1:   --   --   p3  p2
                                        //                                       p3           p2
                                        // xmm1:   -- -- -- --  -- -- -- --  aa bb gg rr  aa bb gg rr
        punpcklbw   xmm1, xmm0          // B -> W
                                        //                  p3                   p2
                                        // xmm1:   00aa 00bb 00gg 00rr  00aa 00bb 00gg 00rr

        movdqa      xmm3, xmm1          //                                       p3
        punpckhqdq  xmm3, xmm1          // xmm3:   ---- ---- ---- ----  00aa 00bb 00gg 00rr

        punpcklwd   xmm1, xmm0          // W -> D
                                        // xmm1:   000000aa  000000bb  000000gg  000000rr, pixcel2
        punpcklwd   xmm3, xmm0          // W -> D
                                        // xmm3:   000000aa  000000bb  000000gg  000000rr, pixcel3

                                        // Pixel 2==========
        cvtdq2ps    xmm1, xmm1          // Dword -> Float
        mulps       xmm1, xmm7          // mono= B * 0.114478f
                                        //          + G * 0.586611f
                                        //              + R * 0.298912f);
                                        // xmm1:   00  B2  G2  R2, pixcel2(float)

                                        // Pixel 3==========
        cvtdq2ps    xmm3, xmm3          // Dword -> Float
        mulps       xmm3, xmm7          // mono= B * 0.114478f
                                        //          + G * 0.586611f
                                        //              + R * 0.298912f);
                                        // xmm3:   00  B3  G3  R3, pixcel3(float)


                                        // pixel 3-2̏
        haddps      xmm1, xmm3          // G + R
                                        // xmm1:   ---B3---  G3+R3   ---B2---  G2+R2  ,  (float)

        haddps      xmm1, xmm1          // G + R + B
                                        // xmm1:   B3+G3+R3 B2+G2+R2 B3+G3+R3 B2+G2+R2,  (float)

        cvtps2dq    xmm5, xmm1          // float -> Dword
                                        // xmm5:   000000M3 000000M2 000000M3 000000M2,  (DWORD)



        movdqa      xmm1, xmm2          // p1, p0=================================
                                        // xmm1:   p3   p2   p1  p0
                                        //                                       p1           p0
                                        // xmm1:   -- -- -- --  -- -- -- --  aa gg bb rr  aa gg bb rr
        punpcklbw   xmm1, xmm0          // B -> W
                                        //                  p1                   p0
                                        // xmm1:   00aa 00gg 00bb 00rr  00aa 00gg 00bb 00rr
                                        //
                                        //                                       p1
        movdqa      xmm3, xmm1
        punpckhqdq  xmm3, xmm1          // xmm3:   ---- ---- ---- ----  00aa 00gg 00bb 00rr
                                        //
        punpcklwd   xmm1, xmm0          // W -> D
                                        // xmm1:   000000aa  000000gg  000000bb  000000rr, pixcel0
        punpcklwd   xmm3, xmm0          // W -> D
                                        // xmm3:   000000aa  000000gg  000000bb  000000rr, pixcel1
                                        

                                        // Pixel 0==========
        cvtdq2ps    xmm1, xmm1          // Dword -> Float
        mulps       xmm1, xmm7          // mono= B * 0.114478f
                                        //          + G * 0.586611f
                                        //              + R * 0.298912f);
                                        // xmm1:   00  B0  G0  R0, pixcel0(float)


                                        // Pixel 1==========
        cvtdq2ps    xmm3, xmm3          // Dword -> Float
        mulps       xmm3, xmm7          // mono= B * 0.114478f
                                        //          + G * 0.586611f
                                        //              + R * 0.298912f);
                                        // xmm3:   00  B1  G1  R1, pixcel1(float)


                                        // pixel 1-0̏
        haddps      xmm1, xmm3          // G + R
                                        // xmm1:   ---B1---  G1+R1   ---B0---  G0+R0  ,  (float)
        haddps      xmm1, xmm1          // G + R + B
                                        // xmm1:   B1+G1+R1 B0+G0+R0 B1+G1+R1 B0+G0+R0,  (float)

        cvtps2dq    xmm1, xmm1          // float -> Dword
                                        // xmm5:   000000M1 000000M0 000000M1 000000M0,  (DWORD)


        movsd       xmm5, xmm1          //  4 pixel𓝍
                                        // xmm5:   000000M3 000000M2 000000M1 000000M0,  (DWORD)

        movdqa      xmm1, xmm5
        pslldq      xmm1, 1             // xmm1:   0000M300 0000M200 0000M100 0000M000,  (DWORD)
        por         xmm1, xmm5          // xmm1:   0000M3M3 0000M2M2 0000M1M1 0000M0M0,  (DWORD)

        movdqa      xmm5, xmm1          // xmm5:   0000M3M3 0000M2M2 0000M1M1 0000M0M0,  (DWORD)
        pslldq      xmm1, 1             // xmm1:   00M3M300 00M2M200 00M1M100 00M0M000,  (DWORD)
        por         xmm5, xmm1          // xmm5:   00M3M3M3 00M2M2M2 00M1M1M1 00M0M0M0,  (DWORD)


        movdqa  [edi+ebx], xmm5         // write to dest
        lea     ebx,  [ebx+16]          // next address

        dec     ecx
        jecxz   exit_x
        jmp     loopLbl
    exit_x:
    }
}